import cv2
import numpy as np
from PIL import Image
from tflite_runtime.interpreter import Interpreter

class Inferencer:
    def __init__(self, model_path, threshold, frame_scale):
        self.interpreter = Interpreter(model_path)
        self.interpreter.allocate_tensors()
        self.input_details = self.interpreter.get_input_details()
        self.output_details = self.interpreter.get_output_details()
        _, self.input_height, self.input_width, _ = self.input_details[0]['shape']
        self.threshold = threshold
        self.frame_width, self.frame_height = frame_scale


    # 推論モデルに画像の前処理を行ってから、画像を入力する
    def _preprocess(self, frame):
        input_data = Image.fromarray(frame).resize((self.input_width, self.input_height), Image.BICUBIC)
        input_data = np.asarray(input_data)
        input_data = np.expand_dims(input_data, axis=0)

        self.interpreter.set_tensor(self.input_details[0]['index'], input_data.astype(np.uint8))


    # 得られたBounding Boxをimshowするサイズに合わせる
    def _format_box(self, box):
        y1, x1, y2, x2 = box
        x1 = int(x1 * (self.frame_width / self.input_width))
        x2 = int(x2 * (self.frame_width / self.input_width))
        y1 = int(y1 * (self.frame_height / self.input_height))
        y2 = int(y2 * (self.frame_height / self.input_height))
        return (y1, x1, y2, x2)


    # 推論を実行する
    def inference(self, frame):
        # 画像の前処理
        input_image = self._preprocess(frame)

        # 推論の実行
        self.interpreter.invoke()

        # 推論結果の取得
        scores = self.interpreter.get_tensor(self.output_details[0]['index'])[0]
        boxes = self.interpreter.get_tensor(self.output_details[1]['index'])[0]

        # 辞書型に整形
        results = []
        for i in range(len(scores)):
            score = float(scores[i] / 255)
            if score >= self.threshold:
                res = {}
                res['score'] = score
                res['box'] = self._format_box(boxes[i])
                results.append(res)

        return results


    # 推論結果を画像に描画する
    def draw_results(self, frame, results, value):
        for result in results:
            y1, x1, y2, x2 = result['box']
            # score = result['score']
            cv2.rectangle(frame, (x1, y1), (x2, y2), (255, 0, 0), 2)
            if value is not None:
                cv2.putText(
                    frame,
                    str('{:.2f}'.format(value)),
                    (x1, y1-10),
                    cv2.FONT_HERSHEY_PLAIN,
                    1.5,
                    (255, 0, 0),
                    2,
                    cv2.LINE_AA
                )
