import cv2
import numpy as np
import statistics
import json

# 物体検出が完璧でないので、検出範囲からマージン分大きな範囲からメーター読み取りを行う
X1_MARGIN = 50
X2_MARGIN = 50
Y1_MARGIN = 50
Y2_MARGIN = 50

RESIZE_WIDTH = 320

class Meter:
    def __init__(self, path, debug=False):
        # param_fileに指定したjsonファイルからメーターの特徴を得る
        self._parse_params(path)

        # 使用する変数を初期化
        self.label_diffs = [] # 隣り合うスケール間の差を格納(5と10なら10-5=5)
        self.diffs = [] # 各スケール間の単位角度あたりの値の増加量を格納
        self.angles = [] # メーターの中心と各スケール点がなす角度を格納

        # param_fileから得たメーターの特徴から、各点の角度などを得る
        self._labeling_positions()
        self._calc_labels_angles()
        self._calc_diffs()

        self.debug = debug


    # メーターの読み取りを行う
    def read(self, image, dbox):
        # 物体検出で得たBounding Boxの形に画像を切り抜く
        im = self._clip_image(image, dbox)

        # 円を検出
        circle = self._detect_circle(im)
        if circle is None:
            if self.debug:
                print('circles not found')
            return None

        # マージンを加味して検出した円に外接する形で画像を切り抜く
        h, w = image.shape[:2]
        center_x = int(circle[0] + dbox[0] - X1_MARGIN)
        center_y = int(circle[1] + dbox[1] - Y1_MARGIN)
        radius = int(circle[2])

        x1 = center_x - radius if center_x - radius >= 0 else 0
        x2 = center_x + radius if center_x + radius <= w else w
        y1 = center_y - radius if center_y - radius >= 0 else 0
        y2 = center_y + radius if center_y + radius <= h else h

        image = image[y1:y2, x1:x2]

        if self.debug:
            cv2.imshow('cliped_image', image)

        circle[:2] = center_x, center_y
        try:
            # 切り抜いた画像から針の角度を得る
            ret, angle = self._get_angle(image)
        except ZeroDivisionError as e:
            return None
        if not ret:
            if self.debug:
                print('can not read from meter')
            return None
        angle = angle - 180

        # 得た針の角度が、アナログメーターが表示しきれない範囲であればreturn
        if angle < self.angles[0] or angle > self.angles[-1]:
            if self.debug:
                print('out of angle')
            return None

        # 角度から、針がメーターのどこを指しているか計算
        group = self._determine_group(angle)
        diff = self.diffs[group]
        return (diff * (angle - self.angles[0])) + self.labels[0]


    def _parse_params(self, path):
        with open(path) as f:
            params = json.load(f)
        self.center = params['center']
        self.positions = params['positions']
        self.labels = params['labels']


    def _calc_angle(self, a, b):
        a = np.array(a)
        b = np.array(b)
        vec = b - a
        return np.rad2deg(np.arctan2(vec[0], -vec[1]))


    # 隣り合うスケール間の差を取る
    def _labeling_positions(self):
        for i in range(len(self.labels) - 1):
            self.label_diffs.append(self.labels[i + 1] - self.labels[i])


    # 各スケール間の単位角度あたりの値の増加量を計算する
    def _calc_diffs(self):
        for i in range(len(self.angles) - 1):
            diff = self.angles[i + 1] - self.angles[i]
            self.diffs.append(self.label_diffs[i] / diff)


    # メーターの中心と各スケール点がなす角度を取る
    def _calc_labels_angles(self):
        for pos in self.positions:
            self.angles.append(self._calc_angle(self.center, pos))


    def _determine_group(self, angle):
        for i in range(len(self.angles) - 1):
            if angle >= self.angles[i] and angle < self.angles[i + 1]:
                return i
        if self.debug:
            print('Error')
        exit(1)


    # 画像の幅のみ指定して、縦横比を維持したままリサイズ
    def _scale_to_width(self, image, width):
        h, w = image.shape[:2]
        resize_rate = width / w
        height = round(h * resize_rate)
        dst = cv2.resize(image, dsize=(width, height))
        return dst, resize_rate


    # 円を検出する
    def _detect_circle(self, image):
        image, resize_rate = self._scale_to_width(image, RESIZE_WIDTH)
        gray = cv2.cvtColor(image, cv2.COLOR_BGR2GRAY)
        gray = cv2.medianBlur(gray, 5)
        circles = cv2.HoughCircles(
            gray,
            cv2.HOUGH_GRADIENT,
            dp=1,
            minDist=1,
            param1=85,
            param2=85,
            minRadius=int(RESIZE_WIDTH/3.5),
            maxRadius=0
        )
        if circles is None:
            return None
        circles = np.uint16(np.around(circles))
        circle = circles[0][0]
        if self.debug:
            cv2.circle(image, (circle[0], circle[1]), circle[2], (0, 255, 0), 3)
            cv2.imshow('circles', image)
        for i in range(len(circle)):
            circle[i] = int(circle[i] / resize_rate)
        return circle


    # 点aと点bを通る直線と、円の中心間の距離を計算
    def _calc_distance(self, a, b, circle):
        x1, y1 = a
        x2, y2 = b
        x3, y3 = circle[:2]
        u = np.array([x2 - x1, y2 - y1])
        v = np.array([x3 - x1, y3 - y1])
        l = abs(np.cross(u, v) / np.linalg.norm(u))
        return l


    # 針の形をなす2本の直線を検出して、針が指す角度を認識する
    def _get_angle(self, image):
        # 再度円を検出して、メーターの中心を確定させる
        circle = self._detect_circle(image)
        if circle is None:
            return (False, 0)

        # 直線検出前に画像からエッジ検出する
        edges = cv2.Canny(image, 50, 200, apertureSize=3)
        theta_t = []
        aa = []
        bb = []

        # 直線検出
        lines = cv2.HoughLines(edges, 1, np.pi/180, 30)

        for i in range(len(lines)):
            for rho, theta in lines[i]:
                a = np.cos(theta)
                b = np.sin(theta)
                x0 = a * rho
                y0 = b * rho
                x1 = int(x0 + 1000 * (-b))
                y1 = int(y0 + 1000 * (a))
                x2 = int(x0 - 1000 * (-b))
                y2 = int(y0 - 1000 * (a))

                # 針をなす直線がほしいので、検出した直線と中心間の距離を計算し、ある程度小さいもののみ使用する
                l = self._calc_distance((x1, x2), (y1, y2), circle)
                if l >= circle[2] / 10:
                    continue

                theta_t.append(theta)
                if self.debug:
                    cv2.line(image,(x1,y1),(x2,y2),(0,0,255),2)

                a0 = (y2 - y1) / (x2 - x1)
                b0 = y1 - (y2 - y1) / (x2 - x1) * x1
                aa.append(a0)
                bb.append(b0)
            if len(aa) >= 2:
                break

        if len(aa) < 2:
            return (False, 0)
        x_t = (bb[1] - bb[0]) / (aa[0] - aa[1])
        y_t = (aa[0] * bb[1] - bb[0] * aa[1]) / (aa[0] - aa[1])

        if x_t < circle[0]:
            theta_hor = statistics.mean(theta_t) * 180 / np.pi
        else:
            theta_hor = 270 - (90 - statistics.mean(theta_t) * 180 / np.pi)

        return (True, theta_hor)


    def _clip_image(self, image, box):
        y1, x1, y2, x2 = box
        h, w = image.shape[:2]
        x1 = x1 - X1_MARGIN if x1 - X1_MARGIN >= 0 else 0
        y1 = y1 - Y1_MARGIN if y1 - Y1_MARGIN >= 0 else 0
        y2 = y2 + X2_MARGIN if y2 + X2_MARGIN <= h else h
        x2 = x2 + Y2_MARGIN if x2 + Y2_MARGIN <= w else w
        return image[y1:y2, x1:x2]
